import os as _os
import tarfile as _tarfile
import zipfile as _zipfile
import urllib as _urllib


def download_file(source='https://www.download.com/file.zip', filepath='./data/file.zip', overwrite=False) -> None:
    """
    Download a file from source location and store it at the specified destination location.

    Args:
        source: the source URL
        filepath: the destination path and filename
        overwrite: if the file already exists then skip the download
    """
    if not overwrite and _os.path.exists(filepath):
        print(f'Download of {source} skipped, file {filepath} exists.')
        return

    path = _os.path.dirname(filepath)
    if path:
        _os.makedirs(path, exist_ok=True)
    else:
        _os.path.join('./', filepath)

    # noinspection PyUnresolvedReferences
    print(f'Download of {source} started.')
    _urllib.request.urlretrieve(source, filepath)
    print(f'Download of {source} completed.')


def extract_tgz(filepath='./data/file.tar.gz', folder='./data') -> None:
    """
    Extracts a .tgz file in the specified destination.

    Args:
        filepath: the .tgz file
        folder: the destination location, if not specified then same path as source file.
    """
    if folder is None:
        folder = _os.path.dirname(filepath)

    print(f"Extraction of {filepath} started.")
    with _tarfile.open(name=filepath) as tObject:
        tObject.extractall(path=folder)

    print(f"Extraction of {filepath} completed.")


def extract_zip(filepath='./data/file.zip', folder='./data') -> None:
    """
    Extracts a .zip file in the specified destination.

    Args:
        filepath: the path and filename
        folder: the destination location, if not specified then same path as source file.
    """
    if folder is None:
        folder = _os.path.dirname(filepath)

    print(f"Extraction of {filepath} started.")
    with _zipfile.ZipFile(file=filepath) as zObject:
        zObject.extractall(path=folder)

    print(f"Extraction of {filepath} completed.")


def list_dir_summary(folder='./data') -> None:
    """
    Lists a summary of the specified folder

    Args:
        folder: the folder to list the contents of
    """
    for dirpath, dirnames, filenames in _os.walk(folder):
        print(f"There are {len(dirnames)} directories, and {len(filenames)} in {dirpath}")



import matplotlib.pyplot as _plt
import seaborn as _sns
import pandas as _pd


def count_categories(dataframe: _pd.DataFrame, label_column: str = None, column_names: list[str] = None, cols=3, figsize: tuple = None):
    if column_names is None:
        column_names = dataframe.select_dtypes(exclude='number').columns.tolist()

    rows = max(int(len(column_names) / cols), 1)
    cols = min(cols, len(column_names))
    rows += 1 if rows * cols < len(column_names) else 0

    # If figsize is not specified then calculate the fig-size
    if figsize is None:
        figsize = (17, rows * 4)

    fig, axs = _plt.subplots(nrows=rows, ncols=cols, figsize=figsize)

    # If we have more then one column then flatten the axis so we can loop through them,
    # if we have only one column then create list containing the axis so we can still loop through it.
    if len(column_names) > 1:
        axs = axs.flatten()
    else:
        axs = [axs]

    n = 0
    for column_name in column_names:
        _sns.countplot(ax=axs[n], data=dataframe, x=column_name, hue=label_column)
        n += 1

    _plt.show()


def histogram_for_columns(dataframe: _pd.DataFrame, column_names: list[str] = None, log=False,
                                  min_nunique: int = 0, max_nunique: int = 50, figsize: tuple = None, cols=3,
                                  verbose=1):
    """
    Plots a histogram for each of the numeric columns in the DataFrame.

    :param dataframe: the pandas dataframe
    :param column_names: columns which exist within the DataFrame if none specified all columns that are numeric
        will be processed
    :param log: set to True to enable logarithmic scaling
    :param min_nunique: minimum number of unique values present, if lower then this then no graph will be displayed (since it is basically boolean)
    :param max_nunique: maximum number of unique values present, only applicable to object column types since these cannot be binned
    :param figsize: size of the plot, if None specified then one is calculated
    :param cols: number of plots on the horizontal axis
    :param verbose: display messages when columns are not visualized
    """
    # assert column_names is not None, "column_names cannot be None"

    # If the column_names argument is not a list then create a list
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    # If we don't have a list of column names then create a histogram for every column.
    if column_names is not None:
        columns = list(dataframe[column_names].columns)
    else:
        columns = list(dataframe.columns)

    # Calculate the number of rows / columns for the subplot.
    rows = max(int(len(columns) / cols), 1)
    cols = min(cols, len(columns))
    rows += 1 if rows * cols < len(columns) else 0

    # If figsize is not specified then calculate the fig-size
    if figsize is None:
        figsize = (17, rows * 4)

    fig, axs = _plt.subplots(nrows=rows, ncols=cols, figsize=figsize)

    # If we have more then one column then flatten the axis so we can loop through them,
    # if we have only one column then create list containing the axis so we can still loop through it.
    if len(columns) > 1:
        axs = axs.flatten()
    else:
        axs = [axs]

    # Horizontal / vertical padding between the histograms.
    fig.tight_layout(h_pad=10, w_pad=5)

    n = 0
    for c in columns:
        v = dataframe[c]
        nunique = v.nunique()

        if _pd.api.types.is_numeric_dtype(v) and min_nunique < nunique:
            # Make sure we do not get more then 50 bins.
            nunique = 50 if nunique > 50 else nunique
            v.hist(ax=axs[n], bins=nunique + 1, log=log, facecolor='#2ab0ff', edgecolor='#169acf', align='left', linewidth=0.1)

            color = iter(["black", "darkred", "red", "orangered", 'limegreen', 'green', 'darkgreen'])
            axs[n].axvline(v.mean(), label=f'Mean', color=next(color), linestyle='dashed')

            z_score_min = [-1 * v.std() + v.mean(), -2 * v.std() + v.mean(), -3 * v.std() + v.mean()]
            for idx, value in enumerate(z_score_min):
                # Only show when it is not before the graph starts.
                c = next(color)
                if value > v.min():
                    axs[n].axvline(value, label=f'(-{idx+1}σ)', color=c, linestyle='dashed')

            z_score_max = [1 * v.std() + v.mean(), 2 * v.std() + v.mean(), 3 * v.std() + v.mean()]
            for idx, value in enumerate(z_score_max):
                # Only show when it is not after the graph starts.
                c = next(color)
                if value < v.max():
                    axs[n].axvline(value, label=f'({idx+1}σ)', color=c, linestyle='dashed')

            axs[n].legend()
        else:
            if verbose:
                print(f"Column '{v.name}' is not visualized, the number of nunique values ({nunique}) either exceeds {max_nunique} or is lower then {min_nunique}.")
            continue

        axs[n].set(title=v.name)

        # Only rotate the x labels
        for tick in axs[n].get_xticklabels():
            tick.set_rotation(45)

        # Remove unnecessary white space on the left/right side of the graph.
        axs[n].margins(x=0)
        axs[n].grid(axis='x')
        n += 1

    for i in range(n, rows*cols):
        fig.delaxes(axs[i])

    _plt.show()



import pandas as _pd
import sweetviz as _sv
from ydata_profiling import ProfileReport as _pr


def profile_report(dataframe: _pd.DataFrame) -> _pr:
    """
    Args:
        dataframe:
    """
    return _pr(df=dataframe)


def sweetvis_report(dataframe: _pd, target_feat: str = None):
    report = _sv.analyze(dataframe, target_feat=target_feat)
    report.show_html()



import numpy as _np


def normalize_vector(x: _np.array):
    """
    Normalizes an array containing vectors of x/y coordinates so that the array does not contain
    negative values.

    :param x: the vector containing values from -X to +X which need to be normalized between 0 and 1
    :return: the normalized vector.
    """
    x = x + (_np.abs(_np.min(x[:, 0])))
    x = x / _np.max(x[:, 0])
    x = x + (_np.abs(_np.min(x[:, 1])))
    return x / _np.max(x[:, 1])



import pandas as _pd
import tensorflow as _tf
import sklearn as _sklearn
import so_ml_tools as _soml


def determine_outliers_for_multiclass_classification(x, y_true, y_pred=None, y_prob=None, top=10) -> _pd.DataFrame:
    """
    Displays the top X (default 10) outliers between y_true and y_pred.

    Args:
        x: data to display
        y_true: actual labels [3, 3, 2, ..., 4, 4, 1]
        y_pred: predictions [3, 3, 2, ..., 4, 2, 1]
        y_prob: probabilities (multiclass) [[0.30974227, ...], [0.32494593, ...], ..] this field is optional, if not provided the y_prob will be determined by y_pred.
        top: how many outliers to return

    Returns:
        The outliers
    """
    # Check if we need to convert y_prob to numpy.
    _y_prob = y_prob
    if _tf.is_tensor(x=y_prob):
        _y_prob = y_prob.numpy()

    # If y_pred is None then determine the predictions, otherwise check if we need to convert to numpy.
    _y_pred = y_pred
    if y_pred is None:
        _y_pred = _soml.util.label.to_prediction(y_prob=_y_prob)
    elif _tf.is_tensor(x=y_pred):
        _y_pred = y_pred.numpy()

    # Create a matrix containing the y_true, y_pred and y_prob value in columns.
    data = None
    if _soml.util.label.is_multiclass_classification(_y_prob):
        data = [[y_true[i], x, _y_prob[i][x]] for i, x in enumerate(_y_pred)]
    elif _soml.util.label.is_binary_classification(_y_prob):
        data = [[y_true[i], x[0], _y_prob[i][0]] for i, x in enumerate(_y_pred)]

    if data is None:
        raise ValueError('y_prob is not a multiclass or binary classification.')

    # Create a dataframe from the matrix with the appropriate column names
    data_df = _pd.DataFrame(data, columns=['y_true', 'y_pred', 'y_prob'])

    # Add the columns of X to the dataframe.
    all_df = data_df
    if x is not None:
        all_df = _pd.concat([x, data_df], axis=1)

    # From all rows that have in-equal y_true vs y_pred, sort on y_prob in descending and select the :top rows.
    outliers = all_df[all_df['y_true'] != all_df['y_pred']].sort_values('y_prob', ascending=False)[:top]

    # Round the y_prob column to 2 decimals, for that we need to cast the y_prob column to float instead of float16
    outliers['y_prob'] = outliers['y_prob'].astype(float).round(decimals=2)

    return outliers


def determine_outliers_for_binary_classification(x, y_true, y_pred, target_column: str, top=10):
    """
    Displays the top X (default 10) outliers between y_true and y_pred.
    :param x: data to display
    :param y_true: labels
    :param y_pred: predictions
    :param target_column: column on which we can calculate the mean negative / positive.
    :param top: how many rows to display
    :return: dataframe containing the top X.
    """
    x_copy = x.copy()
    x_copy["y_true"] = y_true.round(0)
    x_copy["y_pred"] = y_pred.round(0)

    diff = y_true - y_pred
    x_copy["diff"] = diff.round(0)

    mean_positive = diff[diff[target_column] > 0].mean()[0]
    mean_negative = diff[diff[target_column] < 0].mean()[0]

    positive_outliers = x_copy[x_copy["diff"] > mean_positive]
    negative_outliers = x_copy[x_copy["diff"] < mean_negative]

    positive_outliers_sorted = positive_outliers.sort_values(['diff'], ascending=False, inplace=False)[:top]
    negative_outliers_sorted = negative_outliers.sort_values(['diff'], ascending=True, inplace=False)[:top]

    return _pd.concat([positive_outliers_sorted, negative_outliers_sorted])


def classification_report(y_true, y_pred=None, y_prob=None) -> None:
    y_true = _soml.util.label.to_prediction(y_prob=y_true)

    # If y_pred is not supplied but y_prob is then calculatwe y_pred
    if y_pred is None and y_prob is not None:
        y_pred = _soml.util.label.to_prediction(y_prob=y_prob)
    elif y_pred is None and y_prob is None:
        raise "Must specify 'y_pred' or 'y_prob'"

    print(_sklearn.metrics.classification_report(y_true=y_true, y_pred=y_pred, zero_division=0))

# Methods below are less intuitive and informative then using the classification_report.

# def quality_metrics(y_true, y_pred=None, y_prob=None) -> _pd.DataFrame:
#     """
#     calculates model accuracy, precision, recall and F1-Score
#
#     Args:
#         y_true: the truth labels
#         y_pred: (optional) the predictions (either y_pred or y_prob should be supplied)
#         y_prob: (optional) the probabilities (either y_pred or y_prob should be supplied)
#
#     Returns:
#         A 'dict' containing accuracy, precision, recall, f1 score and support
#     """
#     y_true = _soml.util.label.to_prediction(y_prob=y_true)
#
#     # If y_pred is not supplied but y_prob is then calculatwe y_pred
#     if y_pred is None and y_prob is not None:
#         y_pred = _soml.util.label.to_prediction(y_prob=y_prob)
#     elif y_pred is None and y_prob is None:
#         raise "Must specify 'y_pred' or 'y_prob'"
#
#     # Calculate model accuracy
#     model_accuracy = _sklearn.metrics.accuracy_score(y_true, y_pred) * 100
#
#     # Calculate precision, recall and F1-score using "weighted" average,
#     # weighted will also take the amount of samples for each in mind.
#     model_precission, model_recall, model_f1_score, _ = \
#         _sklearn.metrics.precision_recall_fscore_support(y_true, y_pred, average="weighted")
#     model_results = \
#         {
#             "accuracy": model_accuracy,
#             "precision": model_precission,
#             "recall": model_recall,
#             "f1-score": model_f1_score
#         }
#     return _pd.DataFrame(data=dict(sorted(model_results.items())), index=[0])
#
#
# def quality_metrics_diff(metrics_1: _pd.DataFrame, metrics_2: _pd.DataFrame) -> _pd.DataFrame:
#     """
#     Returns the difference between 'metrics_1' and 'metrics_2'.
#
#     Args:
#         metrics_1: the first set of metrics
#         metrics_2: the second set of metrics
#
#     Returns
#         A new 'pd.DataFrame' with the values of 'metrics_1' and 'metrics_2' and difference between these two.
#     """
#     c1 = {'metrics': 'metrics_1'}
#     c1.update(metrics_1.to_dict('records')[0])
#     m1 = _pd.DataFrame(data=c1, index=[0])
#
#     c2 = {'metrics': 'metrics_2'}
#     c2.update(metrics_2.to_dict('records')[0])
#     m2 = _pd.DataFrame(data=c2, index=[0])
#
#     diff = (m1.iloc[0][1:] - m2.iloc[0][1:]).to_frame().transpose()
#     diff.insert(loc=0, column='metrics', value='diff')
#
#     complete = _pd.concat([m1, m2, diff])
#     complete.reset_index(drop=True, inplace=True)
#     return complete
#
#
# def quality_metrics_combine(metrics: dict, sort_by: list[str] = None, ascending: bool | list[bool] | tuple[bool, ...] = False) -> _pd.DataFrame:
#     """
#     Combines all the given metrics into a single overview, call with a dictionary:
#
#         {
#             "baseline": baseline_results,
#             "model_1": model_1_results,
#             "model_2": model_2_results,
#             "model_3": model_3_results
#         }
#
#     Args:
#         metrics: a dictionary of quality metrics pd.DataFrame objects where the key is the name.
#         sort_by: column to sort by, if None specified then the default "accuracy" will be used.
#         ascending: a 'list' or 'tuple' or boolean value containing the sort order.
#
#     Returns:
#         A 'pd.DataFrame' containing all quality metrics combined.
#     """
#     if sort_by is None:
#         sort_by = ["accuracy"]
#
#     all_metrics_results = _pd.concat(metrics)
#     all_metrics_results.reset_index(inplace=True)
#     _soml.pd.dataframe.drop_columns(dataframe=all_metrics_results, column_names=['level_1'])
#     all_metrics_results.rename(columns={"level_0": "name"}, inplace=True)
#     all_metrics_results.sort_values(by=sort_by, inplace=True, ascending=ascending)
#     return all_metrics_results
